# Import all converters at the top to make sure they are registered first
from .converters import *  # noqa: F403, I001

from .compiler import optimize, optimize_single
from .decorators import required_annotations, required_fields
from .fields import DjangoConnectionField, DjangoListField, RelatedField
from .types import DjangoObjectType

__all__ = [
    "DjangoConnectionField",
    "DjangoListField",
    "DjangoObjectType",
    "RelatedField",
    "optimize",
    "optimize_single",
    "required_annotations",
    "required_fields",
]



from __future__ import annotations

from typing import TYPE_CHECKING

import graphene
import graphene_django
from django_filters.constants import ALL_FIELDS
from graphene_django.utils import is_valid_django_model

from .compiler import optimize_single
from .settings import optimizer_settings
from .typing import PK, OptimizedDjangoOptions

if TYPE_CHECKING:
    from django.db.models import Model, QuerySet

    from .typing import Any, GQLInfo, Literal, Optional, TypeVar, Union

    TModel = TypeVar("TModel", bound=Model)


__all__ = [
    "DjangoObjectType",
]


class DjangoObjectType(graphene_django.types.DjangoObjectType):
    """DjangoObjectType that automatically optimizes its queryset."""

    _meta: OptimizedDjangoOptions

    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        _meta: Optional[OptimizedDjangoOptions] = None,
        model: Optional[type[Model]] = None,
        fields: Union[list[str], Literal["__all__"], None] = "__all__",
        max_complexity: Optional[int] = None,
        **options: Any,
    ) -> None:
        if not is_valid_django_model(model):  # pragma: no cover
            msg = f"You need to pass a valid Django Model in {cls.__name__}.Meta, received {model}."
            raise TypeError(msg)

        if _meta is None:
            _meta = OptimizedDjangoOptions(cls)

        if not hasattr(cls, "pk") and (fields == ALL_FIELDS or "pk" in fields):
            cls.pk = graphene.Int() if model._meta.pk.name == "id" else graphene.ID()
            cls.resolve_pk = cls.resolve_id

        _meta.max_complexity = max_complexity or optimizer_settings.MAX_COMPLEXITY
        super().__init_subclass_with_meta__(_meta=_meta, model=model, fields=fields, **options)

    @classmethod
    def filter_queryset(cls, queryset: QuerySet[TModel], info: GQLInfo) -> QuerySet[TModel]:
        """Implement this method filter to the available rows from the model on this node."""
        return queryset

    @classmethod
    def get_queryset(cls, queryset: QuerySet[TModel], info: GQLInfo) -> QuerySet[TModel]:
        return queryset

    @classmethod
    def get_node(cls, info: GQLInfo, pk: PK) -> Optional[TModel]:
        queryset = cls._meta.model._default_manager.all()
        return optimize_single(queryset, info, pk=pk, max_complexity=cls._meta.max_complexity)



# ruff: noqa: UP006
from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

import graphene
from graphene.relay.connection import connection_adapter, page_info_adapter
from graphene.types.argument import to_arguments
from graphene.utils.str_converters import to_snake_case
from graphene_django.settings import graphene_settings
from graphene_django.utils.utils import maybe_queryset
from graphql_relay.connection.array_connection import offset_to_cursor

from .cache import store_in_query_cache
from .compiler import OptimizationCompiler, optimize
from .errors import OptimizerError
from .settings import optimizer_settings
from .utils import calculate_queryset_slice, get_underlying_type, optimizer_logger
from .validators import validate_pagination_args

if TYPE_CHECKING:
    from django.db import models
    from django.db.models.manager import Manager
    from graphene.relay.connection import Connection
    from graphene_django import DjangoObjectType
    from graphql_relay import EdgeType
    from graphql_relay.connection.connection import ConnectionType

    from .filter import FilterSet
    from .typing import (
        Any,
        ConnectionResolver,
        GQLInfo,
        ModelResolver,
        Optional,
        QuerySetResolver,
        Type,
        TypeVar,
        Union,
    )

    TModel = TypeVar("TModel", bound=models.Model)


__all__ = [
    "DjangoConnectionField",
    "DjangoListField",
    "RelatedField",
]


class RelatedField(graphene.Field):
    """Field for `to-one` related models with automatic node resolution."""

    def __init__(self, type_: Union[type[DjangoObjectType], str], *, reverse: bool = False, **kwargs: Any) -> None:
        """
        Initialize a related field for the given type.

        :param type_: Object type or dot import path to the object type.
        :param reverse: Is the relation direction forward or reverse?
        :param kwargs: Extra arguments passed to `graphene.types.field.Field`.
        """
        self.reverse = reverse
        super().__init__(type_, **kwargs)

    def wrap_resolve(self, parent_resolver: ModelResolver) -> ModelResolver:
        if self.reverse:
            return self.reverse_resolver
        return self.forward_resolver

    def forward_resolver(self, root: models.Model, info: GQLInfo) -> Optional[models.Model]:
        field_name = to_snake_case(info.field_name)
        db_field_key: str = root.__class__._meta.get_field(field_name).attname
        object_pk = getattr(root, db_field_key, None)
        if object_pk is None:  # pragma: no cover
            return None
        return self.underlying_type.get_node(info, object_pk)

    def reverse_resolver(self, root: models.Model, info: GQLInfo) -> Optional[models.Model]:
        field_name = to_snake_case(info.field_name)
        # Reverse object should be optimized to the root model.
        reverse_object: Optional[models.Model] = getattr(root, field_name, None)
        if reverse_object is None:  # pragma: no cover
            return None
        return self.underlying_type.get_node(info, reverse_object.pk)

    @cached_property
    def underlying_type(self) -> type[DjangoObjectType]:
        return get_underlying_type(self.type)


class FilteringMixin:
    # Subclasses should implement the following properties:
    underlying_type: type[DjangoObjectType]

    @property
    def args(self) -> dict[str, graphene.Argument]:
        return to_arguments(getattr(self, "_base_args", None) or {}, self.filtering_args)

    @args.setter
    def args(self, args: dict[str, Any]) -> None:
        # noinspection PyAttributeOutsideInit
        self._base_args = args

    @cached_property
    def filterset_class(self) -> Optional[type[FilterSet]]:
        if not self.has_filters:  # pragma: no cover
            return None

        from .filter import get_filterset_for_object_type

        return get_filterset_for_object_type(self.underlying_type)

    @cached_property
    def filtering_args(self) -> Optional[dict[str, graphene.Argument]]:
        if not self.has_filters:  # pragma: no cover
            return None

        from graphene_django.filter.utils import get_filtering_args_from_filterset

        return get_filtering_args_from_filterset(self.filterset_class, self.underlying_type)

    @cached_property
    def has_filters(self) -> bool:
        return bool(self.underlying_type._meta.filter_fields or self.underlying_type._meta.filterset_class)


class DjangoListField(FilteringMixin, graphene.Field):
    """Django list field that also supports filtering."""

    def __init__(self, type_: Union[Type[DjangoObjectType], str], **kwargs: Any) -> None:
        """
        Initialize a list field for the given type.

        :param type_: Object type or dot import path to the object type.
        :param kwargs:  Extra arguments passed to `graphene.types.field.Field`.
        """
        if isinstance(type_, graphene.NonNull):  # pragma: no cover
            type_ = type_.of_type
        super().__init__(graphene.List(graphene.NonNull(type_)), **kwargs)

    def wrap_resolve(self, parent_resolver: QuerySetResolver) -> QuerySetResolver:
        self.resolver = parent_resolver
        return self.list_resolver

    def list_resolver(self, root: Any, info: GQLInfo, **kwargs: Any) -> models.QuerySet:
        result = self.resolver(root, info, **kwargs)
        queryset = self.to_queryset(result)
        queryset = self.underlying_type.get_queryset(queryset, info)

        max_complexity: Optional[int] = getattr(self.underlying_type._meta, "max_complexity", None)
        return optimize(queryset, info, max_complexity=max_complexity)

    def to_queryset(self, iterable: Union[models.QuerySet, Manager, None]) -> models.QuerySet:
        # Default resolver can return a Manager-instance or None.
        if iterable is None:
            iterable = self.model._default_manager
        return maybe_queryset(iterable)

    @cached_property
    def underlying_type(self) -> type[DjangoObjectType]:
        return get_underlying_type(self.type)

    @cached_property
    def model(self) -> type[models.Model]:
        return self.underlying_type._meta.model


class DjangoConnectionField(FilteringMixin, graphene.Field):
    """Connection field for Django models that works for both filtered and non-filtered Relay-nodes."""

    def __init__(self, type_: Union[Type[DjangoObjectType], str], **kwargs: Any) -> None:
        """
        Initialize a connection field for the given type.

        :param type_: DjangoObjectType the connection is for.
        :param kwargs: Extra arguments passed to `graphene.types.field.Field`.
        """
        # Maximum number of items that can be requested in a single query for this connection.
        # Set to None to disable the limit.
        self.max_limit: Optional[int] = kwargs.pop("max_limit", graphene_settings.RELAY_CONNECTION_MAX_LIMIT)

        # Default inputs for a connection field
        kwargs.setdefault("first", graphene.Int())
        kwargs.setdefault("last", graphene.Int())
        kwargs.setdefault("offset", graphene.Int())
        kwargs.setdefault("after", graphene.String())
        kwargs.setdefault("before", graphene.String())

        super().__init__(type_, **kwargs)

    def wrap_resolve(self, parent_resolver: QuerySetResolver) -> ConnectionResolver:
        self.resolver = parent_resolver
        return self.connection_resolver

    def connection_resolver(self, root: Any, info: GQLInfo, **kwargs: Any) -> ConnectionType:
        pagination_args = validate_pagination_args(
            first=kwargs.pop("first", None),
            last=kwargs.pop("last", None),
            offset=kwargs.pop("offset", None),
            after=kwargs.pop("after", None),
            before=kwargs.pop("before", None),
            max_limit=self.max_limit,
        )

        # Call the ObjectType's "resolve_{field_name}" method if it exists.
        # Otherwise, call the default resolver (usually `dict_or_attr_resolver`).
        result = self.resolver(root, info, **kwargs)
        queryset = self.to_queryset(result)
        queryset = pre_optimized_queryset = self.underlying_type.get_queryset(queryset, info)

        max_complexity: Optional[int] = getattr(self.underlying_type._meta, "max_complexity", None)

        try:
            optimizer = OptimizationCompiler(info, max_complexity=max_complexity).compile(queryset)
            if optimizer is not None:
                queryset = optimizer.optimize_queryset(queryset)

            # Queryset optimization contains filtering, so we count after optimization.
            pagination_args["size"] = count = queryset.count()

            # Slice a queryset using the calculated pagination arguments.
            cut = calculate_queryset_slice(**pagination_args)
            queryset = queryset[cut]

            # Store data in cache after pagination
            if optimizer:
                store_in_query_cache(key=info.operation, queryset=queryset, schema=info.schema, optimizer=optimizer)

        except OptimizerError:  # pragma: no cover
            raise

        except Exception as error:  # pragma: no cover  # noqa: BLE001
            if not optimizer_settings.SKIP_OPTIMIZATION_ON_ERROR:
                raise

            # If an error occurs during optimization, we should still return the unoptimized queryset.
            optimizer_logger.info("Something went wrong during the optimization process.", exc_info=error)
            queryset = pre_optimized_queryset
            pagination_args["size"] = count = queryset.count()
            cut = calculate_queryset_slice(**pagination_args)
            queryset = queryset[cut]

        # Create a connection from the sliced queryset.
        edges: list[EdgeType] = [
            self.connection_type.Edge(node=value, cursor=offset_to_cursor(cut.start + index))
            for index, value in enumerate(queryset)
        ]
        connection = connection_adapter(
            cls=self.connection_type,
            edges=edges,
            pageInfo=page_info_adapter(
                startCursor=edges[0].cursor if edges else None,
                endCursor=edges[-1].cursor if edges else None,
                hasPreviousPage=cut.start > 0,
                hasNextPage=cut.stop <= count,
            ),
        )
        connection.iterable = queryset
        connection.length = count
        return connection

    def to_queryset(self, iterable: Union[models.QuerySet, Manager, None]) -> models.QuerySet:
        # Default resolver can return a Manager-instance or None.
        if iterable is None:
            iterable = self.model._default_manager
        return maybe_queryset(iterable)

    @cached_property
    def type(self) -> Union[Type[Connection], graphene.NonNull]:
        type_ = super().type
        non_null = isinstance(type_, graphene.NonNull)
        if non_null:  # pragma: no cover
            type_ = type_.of_type

        connection_type: Optional[Type[Connection]] = type_._meta.connection
        if connection_type is None:  # pragma: no cover
            msg = f"The type {type_.__name__} doesn't have a connection"
            raise ValueError(msg)

        if non_null:  # pragma: no cover
            return graphene.NonNull(connection_type)
        return connection_type

    @cached_property
    def connection_type(self) -> Type[Connection]:
        type_ = self.type
        if isinstance(type_, graphene.NonNull):  # pragma: no cover
            return type_.of_type
        return type_

    @cached_property
    def underlying_type(self) -> Type[DjangoObjectType]:
        return self.connection_type._meta.node

    @cached_property
    def model(self) -> Type[models.Model]:
        return self.underlying_type._meta.model



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .typing import Any, Callable, TypeVar

    TCallable = TypeVar("TCallable", bound=Callable)


__all__ = [
    "required_fields",
    "required_annotations",
]


def required_fields(*args: str) -> Callable[[TCallable], TCallable]:
    """
    Add hints to a resolver to require given fields
    in relation to its DjangoObjectType model.

    :param args: Fields that the decorated resolver needs.
                 Related entity fields can also be used with
                 the field lookup syntax (e.g., 'related__field')
    """

    def decorator(resolver: TCallable) -> TCallable:
        resolver.fields = args  # type: ignore[attr-defined]
        return resolver

    return decorator


def required_annotations(**kwargs: Any) -> Callable[[TCallable], TCallable]:
    """
    Add hints to a resolver function indicating that the given annotations
    should be applied to the ObjectType queryset _after_ filters are applied.
    See. https://docs.djangoproject.com/en/dev/topics/db/aggregation/#order-of-annotate-and-filter-clauses

    :param kwargs: Annotations that the decorated resolver needs.
                   Values should be Expression or F-object instances,
                   or any other value that works with queryset.annotate().
    """

    def decorator(resolver: TCallable) -> TCallable:
        resolver.annotations = kwargs  # type: ignore[attr-defined]
        return resolver

    return decorator



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

from django.core.exceptions import FieldDoesNotExist, ValidationError
from django.db import models
from django.db.models import Expression, Model, Prefetch, QuerySet
from django.db.models.constants import LOOKUP_SEP
from django.db.models.functions import RowNumber
from graphene_django.registry import get_global_registry
from graphene_django.settings import graphene_settings

from .settings import optimizer_settings
from .utils import calculate_queryset_slice, get_filter_info, mark_optimized, optimizer_logger
from .validators import PaginationArgs

if TYPE_CHECKING:
    from .types import DjangoObjectType
    from .typing import Any, GQLInfo, GraphQLFilterInfo, Optional, TypeVar

    TModel = TypeVar("TModel", bound=Model)


__all__ = [
    "QueryOptimizer",
]


@dataclasses.dataclass
class CompilationResults:
    only_fields: list[str] = dataclasses.field(default_factory=list)
    select_related: list[str] = dataclasses.field(default_factory=list)
    prefetch_related: list[Prefetch] = dataclasses.field(default_factory=list)


class QueryOptimizer:
    """Creates optimized queryset based on the optimization data found by the OptimizationCompiler."""

    def __init__(self, model: type[Model], info: GQLInfo) -> None:
        self.model = model
        self.info = info
        self.only_fields: list[str] = []
        self.related_fields: list[str] = []
        self.annotations: dict[str, Expression] = {}
        self.select_related: dict[str, QueryOptimizer] = {}
        self.prefetch_related: dict[str, QueryOptimizer] = {}

    def optimize_queryset(
        self,
        queryset: QuerySet[TModel],
        *,
        filter_info: Optional[GraphQLFilterInfo] = None,
    ) -> QuerySet[TModel]:
        """
        Add the optimizations in this optimizer to the given queryset.

        :param queryset: QuerySet to optimize.
        :param filter_info: Additional filtering info to use for the optimization.
        """
        if filter_info is None:
            filter_info = get_filter_info(self.info)

        results = self.compile(filter_info=filter_info)

        queryset = self.get_filtered_queryset(queryset)

        if filter_info is not None and filter_info.get("filterset_class") is not None:
            filterset = filter_info["filterset_class"](
                data=self.process_filters(filter_info["filters"]),
                queryset=queryset,
                request=self.info.context,
            )
            if not filterset.is_valid():  # pragma: no cover
                raise ValidationError(filterset.form.errors.as_json())

            queryset = filterset.qs

        if results.prefetch_related:
            queryset = queryset.prefetch_related(*results.prefetch_related)
        if results.select_related:
            queryset = queryset.select_related(*results.select_related)
        if not optimizer_settings.DISABLE_ONLY_FIELDS_OPTIMIZATION and (results.only_fields or self.related_fields):
            queryset = queryset.only(*results.only_fields, *self.related_fields)
        if self.annotations:
            queryset = queryset.annotate(**self.annotations)

        mark_optimized(queryset)
        return queryset

    def compile(self, *, filter_info: GraphQLFilterInfo) -> CompilationResults:
        results = CompilationResults(only_fields=self.only_fields.copy())

        for name, optimizer in self.select_related.items():
            # Promote select related to prefetch related if any annotations are needed.
            if optimizer.annotations:
                self.compile_prefetch(name, optimizer, results, filter_info)
            else:
                self.compile_select(name, optimizer, results, filter_info)

        for name, optimizer in self.prefetch_related.items():
            self.compile_prefetch(name, optimizer, results, filter_info)

        return results

    def compile_select(
        self,
        name: str,
        optimizer: QueryOptimizer,
        results: CompilationResults,
        filter_info: GraphQLFilterInfo,
    ) -> None:
        results.select_related.append(name)
        nested_results = optimizer.compile(filter_info=filter_info)
        results.only_fields.extend(f"{name}{LOOKUP_SEP}{only}" for only in nested_results.only_fields)
        results.select_related.extend(f"{name}{LOOKUP_SEP}{select}" for select in nested_results.select_related)
        for prefetch in nested_results.prefetch_related:
            prefetch.add_prefix(name)
            results.prefetch_related.append(prefetch)

    def compile_prefetch(
        self,
        name: str,
        optimizer: QueryOptimizer,
        results: CompilationResults,
        filter_info: GraphQLFilterInfo,
    ) -> None:
        filter_info = filter_info.get("children", {}).get(name, {})
        queryset = self.get_prefetch_queryset(name, optimizer.model, filter_info=filter_info)
        optimized_queryset = optimizer.optimize_queryset(queryset, filter_info=filter_info)
        results.prefetch_related.append(Prefetch(name, optimized_queryset))

    def get_prefetch_queryset(self, name: str, model: type[TModel], filter_info: GraphQLFilterInfo) -> QuerySet[TModel]:
        queryset = model._default_manager.all()

        pagination_args = self.get_pagination_args(filter_info=filter_info)
        # If no pagination arguments are given, then don't limit the nested items (e.g. regular list fields)
        if all(value is None for value in pagination_args.values()):
            return queryset

        # Just use the relay pagination max limit (ignore ConnectionField max limit) for limiting nested items.
        # However, if the limit is se to None, then don't limit the nested items.
        pagination_args["size"] = graphene_settings.RELAY_CONNECTION_MAX_LIMIT
        if pagination_args["size"] is None:  # pragma: no cover
            return queryset

        cut = calculate_queryset_slice(**pagination_args)

        try:
            # Try to find the prefetch join field from the model to use for partitioning.
            field = self.model._meta.get_field(name)
        except FieldDoesNotExist:  # pragma: no cover
            msg = f"Cannot find field {name!r} on model {self.model.__name__!r}. Cannot optimize nested pagination."
            optimizer_logger.warning(msg)
            return queryset

        field_name: str = field.remote_field.attname
        order_by: Optional[list[str]] = (
            # Use the `order_by` from the filter info, if available
            [x for x in filter_info.get("filters", {}).get("order_by", "").split(",") if x]
            # Use the model's `Meta.ordering` if no `order_by` is given
            or model._meta.ordering
            # No ordering if neither is available
            or None
        )

        return (
            # Add a row number to the queryset, and limit the rows for each
            # partition to based on the given pagination arguments.
            queryset.alias(
                _row_number=models.Window(
                    expression=RowNumber(),
                    partition_by=models.F(field_name),
                    order_by=order_by,
                )
            ).filter(_row_number__gte=cut.start, _row_number__lte=cut.stop)
        )

    def get_filtered_queryset(self, queryset: QuerySet[TModel]) -> QuerySet[TModel]:
        object_type: Optional[DjangoObjectType] = get_global_registry().get_type_for_model(queryset.model)
        if callable(getattr(object_type, "filter_queryset", None)):
            return object_type.filter_queryset(queryset, self.info)  # type: ignore[union-attr]
        return queryset  # pragma: no cover

    def process_filters(self, input_data: dict[str, Any]) -> dict[str, Any]:
        from graphene_django.filter.fields import convert_enum

        return {key: convert_enum(value) for key, value in input_data.items()}

    def get_pagination_args(self, filter_info: GraphQLFilterInfo) -> Optional[PaginationArgs]:
        return PaginationArgs(
            after=filter_info.get("filters", {}).get("after"),
            before=filter_info.get("filters", {}).get("before"),
            first=filter_info.get("filters", {}).get("first"),
            last=filter_info.get("filters", {}).get("last"),
            size=None,
        )

    def __add__(self, other: QueryOptimizer) -> QueryOptimizer:
        self.only_fields += other.only_fields
        self.related_fields += other.related_fields
        self.annotations.update(other.annotations)
        self.select_related.update(other.select_related)
        self.prefetch_related.update(other.prefetch_related)
        return self

    def __str__(self) -> str:
        filter_info = get_filter_info(self.info)
        results = self.compile(filter_info=filter_info)
        only = ",".join(results.only_fields)
        select = ",".join(results.select_related)
        prefetch = ",".join(item.prefetch_to for item in results.prefetch_related)
        return f"{only=}|{select=}|{prefetch=}"



from __future__ import annotations

from typing import TYPE_CHECKING

import django_filters
from graphene_django.filter.utils import get_filterset_class
from graphene_django.utils import maybe_queryset

if TYPE_CHECKING:
    from django.db import models
    from graphene_django import DjangoObjectType

    from .typing import Any, FilterFields, Optional

__all__ = [
    "FilterSet",
    "get_filterset_for_object_type",
]


class FilterSet(django_filters.FilterSet):
    @property
    def qs(self) -> models.QuerySet:
        if not hasattr(self, "_qs"):
            # Override the default `qs` property to include this.
            # It's needed to ensure that the `queryset._result_cache`
            # is not cleared when using `queryset.all()`
            qs = maybe_queryset(self.queryset)
            if self.is_bound:
                # ensure form validation before filtering
                # noinspection PyStatementEffect
                self.errors  # noqa: B018
                qs = self.filter_queryset(qs)
            # noinspection PyAttributeOutsideInit
            self._qs = qs
        return self._qs


def get_filterset_for_object_type(object_type: type[DjangoObjectType]) -> type[FilterSet]:
    meta: dict[str, Any] = {
        "model": object_type._meta.model,
        "fields": object_type._meta.filter_fields,
        "filterset_base_class": FilterSet,
    }
    return get_filterset_class(object_type._meta.filterset_class, **meta)


def get_filterset_for_model(
    model: type[models.Model],
    filterset: Optional[type[FilterSet]],
    fields: Optional[FilterFields],
) -> type[FilterSet]:
    meta: dict[str, Any] = {
        "model": model,
        "fields": fields,
        "filterset_base_class": FilterSet,
    }
    return get_filterset_class(filterset, **meta)



from __future__ import annotations

from graphene_django.settings import graphene_settings
from graphql_relay import cursor_to_offset

from .typing import Optional, TypedDict

__all__ = [
    "validate_pagination_args",
]


class PaginationArgs(TypedDict):
    after: Optional[int]
    before: Optional[int]
    first: Optional[int]
    last: Optional[int]
    size: Optional[int]


def validate_pagination_args(  # noqa: C901, PLR0912
    first: Optional[int],
    last: Optional[int],
    offset: Optional[int],
    after: Optional[str],
    before: Optional[str],
    max_limit: Optional[int] = None,
) -> PaginationArgs:
    """
    Validate the pagination arguments and return a dictionary with the validated values.

    :param first: Number of records to return from the beginning.
    :param last: Number of records to return from the end.
    :param offset: Number of records to skip from the beginning.
    :param after: Cursor value for the last record in the previous page.
    :param before: Cursor value for the first record in the next page.
    :param max_limit: Maximum limit for the number of records that can be requested.
    :raises ValueError: Validation error.
    """
    after = cursor_to_offset(after) if after is not None else None
    before = cursor_to_offset(before) if before is not None else None

    if graphene_settings.RELAY_CONNECTION_ENFORCE_FIRST_OR_LAST and not (first or last):  # pragma: no cover
        msg = "You must provide a `first` or `last` for pagination."
        raise ValueError(msg)

    if first is not None:
        if not isinstance(first, int) or first <= 0:
            msg = "Argument 'first' must be a positive integer."
            raise ValueError(msg)

        if isinstance(max_limit, int) and first > max_limit:
            msg = f"Requesting first {first} records exceeds the limit of {max_limit}."
            raise ValueError(msg)

    if last is not None:
        if not isinstance(last, int) or last <= 0:
            msg = "Argument 'last' must be a positive integer."
            raise ValueError(msg)

        if isinstance(max_limit, int) and last > max_limit:
            msg = f"Requesting last {last} records exceeds the limit of {max_limit}."
            raise ValueError(msg)

    if isinstance(max_limit, int) and first is None and last is None:
        first = max_limit

    if offset is not None:
        if after is not None or before is not None:
            msg = "Can only use either `offset` or `before`/`after` for pagination."
            raise ValueError(msg)
        if not isinstance(offset, int) or offset <= 0:
            msg = "Argument `offset` must be a positive integer."
            raise ValueError(msg)

        # Convert offset to after cursor value. Note that after cursor dictates
        # a value _after_ which results should be returned, so we need to subtract
        # 1 from the offset to get the correct cursor value.
        after = offset - 1

    if after is not None and (not isinstance(after, int) or after < 0):
        msg = "The node pointed with `after` does not exist."
        raise ValueError(msg)

    if before is not None and (not isinstance(before, int) or before < 0):
        msg = "The node pointed with `before` does not exist."
        raise ValueError(msg)

    if after is not None and before is not None and after >= before:
        msg = "The node pointed with `after` must be before the node pointed with `before`."
        raise ValueError(msg)

    # Since `after` is also exclusive, we need to add 1 to it, so that slicing works correctly.
    if after is not None:
        after = after + 1

    # Size is added later with `queryset.count()`.
    return PaginationArgs(after=after, before=before, first=first, last=last, size=None)



from __future__ import annotations

from typing import TYPE_CHECKING

from django.test.signals import setting_changed  # type: ignore[attr-defined]
from settings_holder import SettingsHolder, reload_settings

from .typing import NamedTuple

if TYPE_CHECKING:
    from .typing import Any, Union


__all__ = [
    "optimizer_settings",
]


SETTING_NAME: str = "GRAPHQL_QUERY_OPTIMIZER"


class DefaultSettings(NamedTuple):
    QUERY_CACHE_KEY: str = "_query_cache"
    """Key to store fetched model instances under in the GraphQL schema extensions."""
    OPTIMIZER_MARK: str = "_optimized"
    """Key used mark if a queryset has been optimized by the query optimizer."""
    DISABLE_ONLY_FIELDS_OPTIMIZATION: bool = False
    """Disable optimizing fetched fields with `queryset.only()`."""
    MAX_COMPLEXITY: int = 10
    """Default max number of 'select_related' and 'prefetch related' joins optimizer is allowed to optimize."""
    SKIP_OPTIMIZATION_ON_ERROR: bool = False
    """If there is an unexpected error, should the optimizer skip optimization (True) or throw an error (False)?"""


DEFAULTS: dict[str, Any] = DefaultSettings()._asdict()
IMPORT_STRINGS: set[Union[bytes, str]] = set()
REMOVED_SETTINGS: set[str] = {
    "PK_CACHE_KEY",
    "DONT_OPTIMIZE_ON_ERROR",
}

optimizer_settings = SettingsHolder(
    setting_name=SETTING_NAME,
    defaults=DEFAULTS,
    import_strings=IMPORT_STRINGS,
    removed_settings=REMOVED_SETTINGS,
)

reload_my_settings = reload_settings(SETTING_NAME, optimizer_settings)
setting_changed.connect(reload_my_settings)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING
from weakref import WeakKeyDictionary

from .settings import optimizer_settings

if TYPE_CHECKING:
    from django.db.models import Model, QuerySet
    from graphql import GraphQLSchema

    from .optimizer import QueryOptimizer
    from .typing import PK, Hashable, Optional, QueryCache, TableName, TypeVar

    TModel = TypeVar("TModel", bound=Model)


__all__ = [
    "get_from_query_cache",
    "get_query_cache",
    "store_in_query_cache",
]


def get_query_cache(key: Hashable, schema: GraphQLSchema) -> QueryCache:
    """
    Get or create a cache for storing model instances.
    Cache is implemented as a WeakKeyDictionary on the given key,
    stored in the given schema object.

    Items in the cache are stored first by their database table, then by
    their selected fields (as determined by QueryOptimizer), and lastly
    by their primary key. The first two levels of the hierarchy are implemented
    as defaultdicts.

    :param key: Any hashable value that is present only for the duration of
                a single request, e.g., 'info.operation'.
    :param schema: The GraphQLSchema object where the cache will exist.
    :return: The cache.
    """
    cache = schema.extensions.setdefault(optimizer_settings.QUERY_CACHE_KEY, WeakKeyDictionary())
    return cache.setdefault(key, defaultdict(lambda: defaultdict(dict)))  # type: ignore[no-any-return]


def get_from_query_cache(
    key: Hashable,
    schema: GraphQLSchema,
    model: type[TModel],
    pk: PK,
    optimizer: QueryOptimizer,
) -> Optional[TModel]:
    """
    Get the given model instance from query cache.

    :param key: Any hashable value that is present only for the duration of
                a single request, e.g., 'info.operation'.
    :param schema: The GraphQLSchema object where the cache exists.
    :param model: The model type to look for.
    :param pk: The primary key of the model instance to look for.
    :param optimizer: The QueryOptimizer describing the fields that
                  should have been fetched on the model instance.
    :return: The Model instance if it exists in the cache, None if not.
    """
    optimizer_key = str(optimizer)
    query_cache = get_query_cache(key, schema)
    return query_cache[model._meta.db_table][optimizer_key].get(pk)


def store_in_query_cache(
    *,
    key: Hashable,
    queryset: QuerySet,
    schema: GraphQLSchema,
    optimizer: QueryOptimizer,
) -> None:
    """
    Set all given models, as well as any related models joined to them
    as described by the given QueryOptimizer, to the query cache.

    :param key: Any hashable value that is present only for the duration of
                a single request, e.g., 'info.operation'.
    :param queryset: QuerySet that should be stored in the query cache.
    :param schema: The GraphQLSchema object where the cache exists.
    :param optimizer: The QueryOptimizer describing the fields that
                      are fetched on the model instances.
    """
    query_cache = get_query_cache(key, schema)
    queryset = list(queryset)  # the database query will occur here
    if not queryset:  # pragma: no cover
        return

    for item in queryset:
        _add_item(query_cache, item, optimizer)


def _add_item(query_cache: QueryCache, instance: Model, optimizer: QueryOptimizer) -> None:
    optimizer_key = str(optimizer)
    table_name: TableName = instance._meta.db_table
    query_cache[table_name][optimizer_key][instance.pk] = instance
    _add_selected(query_cache, instance, optimizer)
    _add_prefetched(query_cache, instance, optimizer)


def _add_selected(query_cache: QueryCache, instance: Model, optimizer: QueryOptimizer) -> None:
    for nested_name, nested_optimizer in optimizer.select_related.items():
        # For forward one-to-one and many-to-one, the relation might be null.
        # For reverse one-to-one, the relation might not exist.
        nested_instance: Optional[Model] = getattr(instance, nested_name, None)
        if nested_instance is not None:
            _add_item(query_cache, nested_instance, nested_optimizer)


def _add_prefetched(query_cache: QueryCache, instance: Model, optimizer: QueryOptimizer) -> None:
    for nested_name, nested_optimizer in optimizer.prefetch_related.items():
        # Here we can fetch the many-related items from the instance with `.all()`
        # without hitting the database, because the items have already been prefetched.
        # See: `django.db.models.fields.related_descriptors.RelatedManager.get_queryset`
        # and `django.db.models.fields.related_descriptors.ManyRelatedManager.get_queryset`
        selected: QuerySet[Model] = getattr(instance, nested_name).all()
        for select in selected:
            _add_item(query_cache, select, nested_optimizer)



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, Optional, Union

from django.core.exceptions import FieldDoesNotExist
from django.db.models import Expression, ForeignKey, Manager, ManyToOneRel, Model, QuerySet
from django.db.models.constants import LOOKUP_SEP
from graphene.relay.connection import ConnectionOptions
from graphene.utils.str_converters import to_snake_case
from graphene_django.types import DjangoObjectTypeOptions
from graphene_django.utils import maybe_queryset
from graphql import FieldNode, FragmentSpreadNode, GraphQLField, GraphQLOutputType, InlineFragmentNode, SelectionNode

from .cache import get_from_query_cache, store_in_query_cache
from .errors import OptimizerError
from .optimizer import QueryOptimizer
from .settings import optimizer_settings
from .utils import (
    get_field_type,
    get_selections,
    get_underlying_type,
    is_foreign_key_id,
    is_optimized,
    is_to_many,
    is_to_one,
    optimizer_logger,
)

if TYPE_CHECKING:
    from graphene.types.definitions import GrapheneObjectType, GrapheneUnionType

    from .typing import PK, GQLInfo, ModelField, ToManyField, ToOneField, TypeOptions, TypeVar

    TModel = TypeVar("TModel", bound=Model)


__all__ = [
    "optimize",
    "optimize_single",
    "OptimizationCompiler",
]


def optimize(
    queryset: QuerySet[TModel],
    info: GQLInfo,
    *,
    max_complexity: Optional[int] = None,
) -> QuerySet[TModel]:
    """Optimize the given queryset according to the field selections received in the GraphQLResolveInfo."""
    try:
        optimizer = OptimizationCompiler(info, max_complexity=max_complexity).compile(queryset)
        if optimizer is None:
            return queryset

        optimized_queryset = optimizer.optimize_queryset(queryset)
        store_in_query_cache(key=info.operation, queryset=optimized_queryset, schema=info.schema, optimizer=optimizer)
        return optimized_queryset  # noqa: TRY300

    except OptimizerError:  # pragma: no cover
        raise

    except Exception as error:  # pragma: no cover  # noqa: BLE001
        if not optimizer_settings.SKIP_OPTIMIZATION_ON_ERROR:
            raise

        optimizer_logger.info("Something went wrong during the optimization process.", exc_info=error)
        return queryset


def optimize_single(
    queryset: QuerySet[TModel],
    info: GQLInfo,
    *,
    pk: PK,
    max_complexity: Optional[int] = None,
) -> Optional[TModel]:
    """Optimize the given queryset for a single model instance by its primary key."""
    queryset = queryset.filter(pk=pk)

    try:
        optimizer = OptimizationCompiler(info, max_complexity=max_complexity).compile(queryset)
        if optimizer is None:  # pragma: no cover
            return queryset.first()

        cached_item = get_from_query_cache(info.operation, info.schema, queryset.model, pk, optimizer)
        if cached_item is not None:
            return cached_item

        optimized_queryset = optimizer.optimize_queryset(queryset)
        store_in_query_cache(key=info.operation, queryset=optimized_queryset, schema=info.schema, optimizer=optimizer)

        # Shouldn't use .first(), as it can apply additional ordering, which would cancel the optimization.
        # The queryset should have the right model instance, since we started by filtering by its pk,
        # so we can just pick that out of the result cache (if it hasn't been filtered out).
        return next(iter(optimized_queryset._result_cache or []), None)

    except OptimizerError:  # pragma: no cover
        raise

    except Exception as error:  # pragma: no cover  # noqa: BLE001
        if not optimizer_settings.SKIP_OPTIMIZATION_ON_ERROR:
            raise

        optimizer_logger.info("Something went wrong during the optimization process.", exc_info=error)
        return queryset.first()


class OptimizationCompiler:
    """Class for walking the GraphQL AST and compiling SQL optimizations for based on the given query."""

    def __init__(self, info: GQLInfo, max_complexity: Optional[int] = None) -> None:
        """
        Initialize the optimization compiler with the query info.

        :param info: The GraphQLResolveInfo containing the query AST.
        :param max_complexity: How many 'select_related' and 'prefetch_related' table joins are allowed.
                               Used to protect from malicious queries.
        """
        self.info = info
        self.max_complexity = max_complexity or optimizer_settings.MAX_COMPLEXITY
        self.current_complexity = 0

    def increase_complexity(self) -> None:
        """Increase the current complexity, and check if it exceeds the maximum allowed."""
        self.current_complexity += 1
        if self.current_complexity > self.max_complexity:
            msg = f"Query complexity exceeds the maximum allowed of {self.max_complexity}"
            raise OptimizerError(msg)

    def compile(self, queryset: Union[QuerySet, Manager]) -> Optional[QueryOptimizer]:
        """
        Compile optimizations for the given queryset.

        :return: QueryOptimizer instance that can perform any needed optimization,
                                or None if queryset is already optimized.
        :raises OptimizerError: Something went wrong during the optimization process.
        """
        queryset = maybe_queryset(queryset)

        # If prior optimization has been done already, return early.
        if is_optimized(queryset):
            return None

        field_type = get_field_type(self.info)
        selections = get_selections(self.info)
        if not selections:  # pragma: no cover
            return None

        # Run the optimization process.
        optimizer = self.handle_selections(field_type, selections, queryset.model)

        # When resolving reverse one-to-many relations (other model has foreign key to this model),
        # if `known_related_fields` exist, they should be added to the optimizer, since they are used to linked
        # to the original model based on that field.
        if queryset._known_related_objects:  # pragma: no cover
            optimizer.related_fields += [row.attname for row in queryset._known_related_objects]

        return optimizer

    def handle_selections(
        self,
        field_type: Union[GrapheneObjectType, GrapheneUnionType],
        selections: tuple[SelectionNode, ...],
        model: type[Model],
    ) -> QueryOptimizer:
        optimizer = QueryOptimizer(model=model, info=self.info)

        for selection in selections:
            if isinstance(selection, FieldNode):
                self.handle_field_node(field_type, selection, optimizer)

            elif isinstance(selection, FragmentSpreadNode):
                self.handle_fragment_spread(field_type, selection, model, optimizer)

            elif isinstance(selection, InlineFragmentNode):
                self.handle_inline_fragment(field_type, selection, model, optimizer)

            else:  # pragma: no cover
                msg = f"Unhandled selection node: '{selection}'"
                raise OptimizerError(msg)

        return optimizer

    def handle_field_node(
        self,
        field_type: GrapheneObjectType,
        selection: FieldNode,
        optimizer: QueryOptimizer,
    ) -> None:
        options: TypeOptions = field_type.graphene_type._meta

        if isinstance(options, ConnectionOptions):
            return self.handle_connection_node(field_type, selection, optimizer)

        if isinstance(options, DjangoObjectTypeOptions):
            return self.handle_regular_node(field_type, selection, optimizer)

        msg = f"Unhandled field options type: {options}"  # pragma: no cover
        raise OptimizerError(msg)  # pragma: no cover

    def handle_regular_node(
        self,
        field_type: GrapheneObjectType,
        selection: FieldNode,
        optimizer: QueryOptimizer,
    ) -> None:
        model: type[Model] = field_type.graphene_type._meta.model
        selection_graphql_name = selection.name.value
        selection_graphql_field = field_type.fields.get(selection_graphql_name)
        if selection_graphql_field is None:  # pragma: no cover
            return

        model_field_name = to_snake_case(selection_graphql_name)
        try:
            if model_field_name == "pk":
                model_field: ModelField = model._meta.pk
                model_field_name = model_field.name  # use actual model pk name, e.g. 'id'
            else:
                model_field: ModelField = model._meta.get_field(model_field_name)
        except FieldDoesNotExist:
            self.check_resolver_hints(selection_graphql_field, model, optimizer)
            return

        if not model_field.is_relation or is_foreign_key_id(model_field, model_field_name):
            optimizer.only_fields.append(model_field_name)

        elif is_to_one(model_field):  # noinspection PyTypeChecker
            self.handle_to_one(model_field_name, selection, selection_graphql_field.type, model_field, optimizer)

        elif is_to_many(model_field):  # noinspection PyTypeChecker
            self.handle_to_many(model_field_name, selection, selection_graphql_field.type, model_field, optimizer)

        else:  # pragma: no cover
            msg = f"Unhandled selection: '{selection.name.value}'"
            raise OptimizerError(msg)

    def handle_connection_node(
        self,
        field_type: GrapheneObjectType,
        selection: FieldNode,
        optimizer: QueryOptimizer,
    ) -> None:
        if selection.selection_set is None:  # pragma: no cover
            return

        node: FieldNode = selection.selection_set.selections[0]  # type: ignore[assignment]
        if node.selection_set is None:  # page info
            return

        edges_field = field_type.fields[selection.name.value]
        edge_type = get_underlying_type(edges_field.type)
        node_field = edge_type.fields[node.name.value]
        node_type = get_underlying_type(node_field.type)
        node_model: type[Model] = node_type.graphene_type._meta.model

        nested_optimizer = self.handle_selections(node_type, node.selection_set.selections, node_model)
        optimizer += nested_optimizer

    def handle_to_one(
        self,
        model_field_name: str,
        selection: FieldNode,
        selection_field_type: GraphQLOutputType,
        model_field: ToOneField,
        optimizer: QueryOptimizer,
    ) -> None:
        if selection.selection_set is None:  # pragma: no cover
            return

        related_model: type[Model] = model_field.related_model  # type: ignore[assignment]
        if related_model == "self":  # pragma: no cover
            related_model = model_field.model

        selection_field_type = get_underlying_type(selection_field_type)

        self.increase_complexity()
        nested_optimizer = self.handle_selections(
            selection_field_type,
            selection.selection_set.selections,
            related_model,
        )

        if isinstance(model_field, ForeignKey):
            optimizer.related_fields.append(model_field.attname)

        optimizer.select_related[model_field_name] = nested_optimizer

    def handle_to_many(
        self,
        model_field_name: str,
        selection: FieldNode,
        selection_field_type: GraphQLOutputType,
        model_field: ToManyField,
        optimizer: QueryOptimizer,
    ) -> None:
        if selection.selection_set is None:  # pragma: no cover
            return

        related_model: type[Model] = model_field.related_model  # type: ignore[assignment]
        if related_model == "self":  # pragma: no cover
            related_model = model_field.model

        selection_field_type = get_underlying_type(selection_field_type)

        self.increase_complexity()
        nested_optimizer = self.handle_selections(
            selection_field_type,
            selection.selection_set.selections,
            related_model,
        )

        if isinstance(model_field, ManyToOneRel):
            nested_optimizer.related_fields.append(model_field.field.attname)

        optimizer.prefetch_related[model_field_name] = nested_optimizer

    def handle_fragment_spread(
        self,
        field_type: GrapheneObjectType,
        selection: FragmentSpreadNode,
        model: type[Model],
        optimizer: QueryOptimizer,
    ) -> None:
        graphql_name = selection.name.value
        field_node = self.info.fragments[graphql_name]
        selections = field_node.selection_set.selections
        nested_optimizer = self.handle_selections(field_type, selections, model)
        optimizer += nested_optimizer

    def handle_inline_fragment(
        self,
        field_type: GrapheneUnionType,
        selection: InlineFragmentNode,
        model: type[Model],
        optimizer: QueryOptimizer,
    ) -> None:
        fragment_type_name = selection.type_condition.name.value
        selection_graphql_field: Optional[GrapheneObjectType]
        selection_graphql_field = next((t for t in field_type.types if t.name == fragment_type_name), None)
        if selection_graphql_field is None:  # pragma: no cover
            return

        fragment_model: type[Model] = selection_graphql_field.graphene_type._meta.model
        if fragment_model != model:  # pragma: no cover
            return

        selections = selection.selection_set.selections
        nested_optimizer = self.handle_selections(selection_graphql_field, selections, fragment_model)
        optimizer += nested_optimizer

    def check_resolver_hints(self, field: GraphQLField, model: type[Model], optimizer: QueryOptimizer) -> None:
        anns: Optional[dict[str, Expression]] = getattr(field.resolve, "annotations", None)
        fields: Optional[tuple[str, ...]] = getattr(field.resolve, "fields", None)

        if anns:
            optimizer.annotations.update(anns)
        if fields is None:  # pragma: no cover
            return

        model_fields: list[ModelField] = model._meta.get_fields()
        for field_name in fields:
            hint_optimizer = QueryOptimizer(model=model, info=self.info)
            self.find_field_from_model(field_name, model_fields, hint_optimizer)
            optimizer += hint_optimizer

    def find_field_from_model(
        self,
        field_name: str,
        model_fields: Iterable[ModelField],
        optimizer: QueryOptimizer,
        prefix: str = "",
    ) -> None:
        for model_field in model_fields:
            model_field_name = model_field.name
            if prefix:
                model_field_name = f"{prefix}{LOOKUP_SEP}{model_field_name}"

            if field_name == model_field_name:
                optimizer.only_fields.append(model_field.name)
                return None

            # Check if the hint is to a related field, and if so, recurse into the related model.
            if not f"{field_name}{LOOKUP_SEP}".startswith(f"{model_field_name}{LOOKUP_SEP}"):
                continue

            related_model: type[Model] = model_field.related_model  # type: ignore[assignment]
            if related_model is None:  # pragma: no cover
                msg = (
                    f"Hint {model_field_name!r} seems to be for a related model,"
                    f"but no related model was not found: {field_name!r}"
                )
                raise OptimizerError(msg)

            if related_model == "self":  # pragma: no cover
                related_model = model_field.model

            self.increase_complexity()
            nested_optimizer = QueryOptimizer(model=related_model, info=self.info)

            if is_to_many(model_field):
                optimizer.prefetch_related[model_field.name] = nested_optimizer
            elif is_to_one(model_field):
                optimizer.select_related[model_field.name] = nested_optimizer
            else:  # pragma: no cover
                msg = f"Field {model_field} is not a related field."
                raise OptimizerError(msg)

            if isinstance(model_field, ManyToOneRel):
                nested_optimizer.related_fields.append(model_field.field.attname)

            related_model_fields: list[ModelField] = related_model._meta.get_fields()

            return self.find_field_from_model(
                field_name=field_name,
                prefix=model_field_name,
                optimizer=nested_optimizer,
                model_fields=related_model_fields,
            )

        msg = f"Field {field_name!r} not found in fields: {model_fields}."  # pragma: no cover
        raise OptimizerError(msg)  # pragma: no cover



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from django.db.models import ForeignKey, Model, QuerySet
from graphene import Connection
from graphene.utils.str_converters import to_snake_case
from graphql import (
    FieldNode,
    GraphQLField,
    GraphQLObjectType,
    GraphQLSchema,
    get_argument_values,
)
from graphql.execution.execute import get_field_def

from .settings import optimizer_settings
from .typing import GraphQLFilterInfo, overload

if TYPE_CHECKING:
    from graphene.types.definitions import GrapheneObjectType
    from graphene_django import DjangoObjectType
    from graphql import GraphQLOutputType, SelectionNode

    from .typing import (
        Any,
        FieldNodes,
        GQLInfo,
        ModelField,
        Optional,
        ParamSpec,
        ToManyField,
        ToOneField,
        TypeGuard,
        TypeVar,
        Union,
    )

    T = TypeVar("T")
    P = ParamSpec("P")


__all__ = [
    "get_field_type",
    "get_filter_info",
    "get_selections",
    "get_underlying_type",
    "is_foreign_key_id",
    "is_optimized",
    "is_to_many",
    "is_to_one",
    "mark_optimized",
    "mark_unoptimized",
    "optimizer_logger",
]


optimizer_logger = logging.getLogger("query_optimizer")


def is_foreign_key_id(model_field: ModelField, name: str) -> bool:
    return isinstance(model_field, ForeignKey) and model_field.name != name and model_field.get_attname() == name


@overload
def get_underlying_type(field_type: type[GraphQLOutputType]) -> type[Union[DjangoObjectType, GrapheneObjectType]]:
    ...  # pragma: no cover


@overload
def get_underlying_type(field_type: GraphQLOutputType) -> Union[DjangoObjectType, GrapheneObjectType]:
    ...  # pragma: no cover


def get_underlying_type(field_type):
    while hasattr(field_type, "of_type"):
        field_type = field_type.of_type
    return field_type


def is_to_many(model_field: ModelField) -> TypeGuard[ToManyField]:
    return bool(model_field.one_to_many or model_field.many_to_many)


def is_to_one(model_field: ModelField) -> TypeGuard[ToOneField]:
    return bool(model_field.many_to_one or model_field.one_to_one)


def get_field_type(info: GQLInfo) -> GrapheneObjectType:
    field_node = info.field_nodes[0]
    field_def = get_field_def(info.schema, info.parent_type, field_node)
    return get_underlying_type(field_def.type)


def get_selections(info: GQLInfo) -> tuple[SelectionNode, ...]:
    field_node = info.field_nodes[0]
    selection_set = field_node.selection_set
    return () if selection_set is None else selection_set.selections


def mark_optimized(queryset: QuerySet) -> None:
    """Mark queryset as optimized so that later optimizers know to skip optimization"""
    queryset._hints[optimizer_settings.OPTIMIZER_MARK] = True  # type: ignore[attr-defined]


def mark_unoptimized(queryset: QuerySet) -> None:  # pragma: no cover
    """Mark queryset as unoptimized so that later optimizers will run optimization"""
    queryset._hints.pop(optimizer_settings.OPTIMIZER_MARK, None)  # type: ignore[attr-defined]


def is_optimized(queryset: QuerySet) -> bool:
    """Has the queryset be optimized?"""
    return queryset._hints.get(optimizer_settings.OPTIMIZER_MARK, False)  # type: ignore[no-any-return, attr-defined]


def calculate_queryset_slice(
    *,
    after: Optional[int],
    before: Optional[int],
    first: Optional[int],
    last: Optional[int],
    size: int,
) -> slice:
    """
    Calculate queryset slicing based on the provided arguments.
    Before this, the arguments should be validated so that:
     - `first` and `last`, positive integers or `None`
     - `after` and `before` are non-negative integers or `None`
     - If both `after` and `before` are given, `after` is less than or equal to `before`

    This function is based on the Relay pagination algorithm.
    See. https://relay.dev/graphql/connections.htm#sec-Pagination-algorithm

    :param after: The index after which to start (exclusive).
    :param before: The index before which to stop (exclusive).
    :param first: The number of items to return from the start.
    :param last: The number of items to return from the end (after evaluating first).
    :param size: The total number of items in the queryset.
    """
    #
    # Start from form fetching max number of items.
    #
    start: int = 0
    stop: int = size
    #
    # If `after` is given, change the start index to `after`.
    # If `after` is greater than the current queryset size, change it to `size`.
    #
    if after is not None:
        start = min(after, stop)
    #
    # If `before` is given, change the stop index to `before`.
    # If `before` is greater than the current queryset size, change it to `size`.
    #
    if before is not None:
        stop = min(before, stop)
    #
    # If first is given, and it's smaller than the current queryset size,
    # change the stop index to `start + first`
    # -> Length becomes that of `first`, and the items after it have been removed.
    #
    if first is not None and first < (stop - start):
        stop = start + first
    #
    # If last is given, and it's smaller than the current queryset size,
    # change the start index to `stop - last`.
    # -> Length becomes that of `last`, and the items before it have been removed.
    #
    if last is not None and last < (stop - start):
        start = stop - last

    return slice(start, stop)


def get_filter_info(info: GQLInfo) -> GraphQLFilterInfo:
    """Find filter arguments from the GraphQL query."""
    args = _get_arguments(info.field_nodes, info.variable_values, info.parent_type, info.schema)
    if not args:
        return {}
    return args[to_snake_case(info.field_name)]


def _get_arguments(
    field_nodes: FieldNodes,
    variable_values: dict[str, Any],
    parent: GraphQLObjectType,
    schema: GraphQLSchema,
) -> dict[str, GraphQLFilterInfo]:
    arguments: dict[str, GraphQLFilterInfo] = {}
    for field_node in field_nodes:
        if not isinstance(field_node, FieldNode):  # pragma: no cover
            continue

        field_def: Optional[GraphQLField] = get_field_def(schema, parent, field_node)
        if field_def is None:  # pragma: no cover
            continue

        name = to_snake_case(field_node.name.value)
        filters = get_argument_values(type_def=field_def, node=field_node, variable_values=variable_values)

        new_parent = get_underlying_type(field_def.type)
        if hasattr(new_parent, "graphene_type") and issubclass(new_parent.graphene_type, Connection):
            # Skip the `edges` and `node` fields
            field_def = new_parent.fields["edges"]
            new_parent = get_underlying_type(field_def.type)
            field_def = new_parent.fields["node"]
            new_parent = get_underlying_type(field_def.type)
            field_node = field_node.selection_set.selections[0].selection_set.selections[0]  # noqa: PLW2901

        arguments[name] = info = GraphQLFilterInfo(
            name=new_parent.name,
            filters=filters,
            children={},
            filterset_class=None,
        )

        if hasattr(new_parent, "graphene_type"):
            filter_fields = getattr(new_parent.graphene_type._meta, "filter_fields", None)
            filterset_class = getattr(new_parent.graphene_type._meta, "filterset_class", None)
            if filterset_class is not None or filter_fields is not None:
                from .filter import get_filterset_for_model

                model: type[Model] = new_parent.graphene_type._meta.model
                info["filterset_class"] = get_filterset_for_model(model, filterset_class, filter_fields)

        if field_node.selection_set is not None:
            result = _get_arguments(field_node.selection_set.selections, variable_values, new_parent, schema)
            if result:
                info["children"] = result

    return {name: field for name, field in arguments.items() if field["filters"] or field["children"]}



from __future__ import annotations

from typing import TYPE_CHECKING

import graphene
from django.db import models
from graphene_django.converter import convert_django_field, get_django_field_description

if TYPE_CHECKING:
    from graphene_django.registry import Registry

    from query_optimizer import DjangoObjectType
    from query_optimizer.fields import DjangoConnectionField, DjangoListField
    from query_optimizer.typing import Optional, Union

__all__ = [
    "convert_to_one_field",
    "convert_to_many_field",
]


@convert_django_field.register(models.OneToOneField)
@convert_django_field.register(models.ForeignKey)
@convert_django_field.register(models.OneToOneRel)
def convert_to_one_field(
    field,  # noqa: ANN001
    registry: Optional[Registry] = None,
) -> graphene.Dynamic:
    def dynamic_type() -> Optional[graphene.Field]:
        type_: Optional[type[DjangoObjectType]] = registry.get_type_for_model(field.related_model)
        if type_ is None:  # pragma: no cover
            return None

        actual_field = field.field if isinstance(field, models.OneToOneRel) else field
        description: str = get_django_field_description(actual_field)
        required: bool = False if isinstance(field, models.OneToOneRel) else not field.null
        reverse: bool = isinstance(field, models.OneToOneRel)

        from query_optimizer.fields import RelatedField

        return RelatedField(
            type_,
            reverse=reverse,
            description=description,
            required=required,
        )

    return graphene.Dynamic(dynamic_type)


@convert_django_field.register(models.ManyToManyField)
@convert_django_field.register(models.ManyToManyRel)
@convert_django_field.register(models.ManyToOneRel)
def convert_to_many_field(
    field,  # noqa: ANN001
    registry: Optional[Registry] = None,
) -> graphene.Dynamic:
    def dynamic_type() -> Union[DjangoConnectionField, DjangoListField, None]:
        type_: Optional[type[DjangoObjectType]] = registry.get_type_for_model(field.related_model)
        if type_ is None:  # pragma: no cover
            return None

        actual_field = field if isinstance(field, models.ManyToManyField) else field.field
        description: str = get_django_field_description(actual_field)
        required: bool = True  # will always return a queryset, even if empty

        from query_optimizer.fields import DjangoConnectionField, DjangoListField

        if type_._meta.connection:  # pragma: no cover
            return DjangoConnectionField(
                type_,
                required=required,
                description=description,
            )
        return DjangoListField(
            type_,
            required=required,
            description=description,
        )

    return graphene.Dynamic(dynamic_type)



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    Generator,
    Hashable,
    Iterable,
    Literal,
    NamedTuple,
    Optional,
    Type,
    TypedDict,
    TypeVar,
    Union,
    overload,
)

from graphene.relay.connection import ConnectionOptions
from graphene_django.types import DjangoObjectTypeOptions
from graphql_relay import ConnectionType

# New in version 3.10
try:
    from typing import ParamSpec, TypeAlias, TypeGuard
except ImportError:
    from typing_extensions import ParamSpec, TypeAlias, TypeGuard


from django.core.handlers.wsgi import WSGIRequest
from django.db.models import (
    Field,
    ForeignKey,
    ForeignObject,
    ForeignObjectRel,
    Manager,
    ManyToManyField,
    ManyToManyRel,
    ManyToOneRel,
    Model,
    OneToOneField,
    QuerySet,
)
from graphql import FieldNode, GraphQLResolveInfo, SelectionNode

if TYPE_CHECKING:
    from django.contrib.auth.models import AnonymousUser, User
    from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
    from django_filters import FilterSet

__all__ = [
    "Any",
    "Callable",
    "Collection",
    "ConnectionResolver",
    "FieldNodes",
    "FilterFields",
    "GQLInfo",
    "Generator",
    "GraphQLFilterInfo",
    "Hashable",
    "Iterable",
    "Literal",
    "ModelField",
    "ModelResolver",
    "NamedTuple",
    "OptimizedDjangoOptions",
    "Optional",
    "PK",
    "ParamSpec",
    "QueryCache",
    "QuerySetResolver",
    "OptimizerKey",
    "TableName",
    "ToManyField",
    "ToOneField",
    "Type",
    "TypeGuard",
    "TypeOptions",
    "TypeVar",
    "TypedDict",
    "Union",
    "overload",
]


TModel = TypeVar("TModel", bound=Model)
TableName: TypeAlias = str
OptimizerKey: TypeAlias = str
PK: TypeAlias = Any
QueryCache: TypeAlias = dict[TableName, dict[OptimizerKey, dict[PK, TModel]]]
ModelField: TypeAlias = Union[Field, ForeignObjectRel, "GenericForeignKey"]
ToManyField: TypeAlias = Union["GenericRelation", ManyToManyField, ManyToOneRel, ManyToManyRel]
ToOneField: TypeAlias = Union["GenericRelation", ForeignObject, ForeignKey, OneToOneField]
TypeOptions: TypeAlias = Union[DjangoObjectTypeOptions, ConnectionOptions]
AnyUser: TypeAlias = Union["User", "AnonymousUser"]
FilterFields: TypeAlias = Union[dict[str, list[str]], list[str]]

QuerySetResolver = Callable[..., Union[QuerySet, Manager, None]]
ModelResolver = Callable[..., Union[Model, None]]
ConnectionResolver = Callable[..., ConnectionType]
FieldNodes = Iterable[Union[FieldNode, SelectionNode]]


class UserHintedWSGIRequest(WSGIRequest):
    user: AnyUser


class GQLInfo(GraphQLResolveInfo):
    context: UserHintedWSGIRequest


class OptimizedDjangoOptions(DjangoObjectTypeOptions):
    max_complexity: int


class GraphQLFilterInfo(TypedDict, total=False):
    name: str
    filters: dict[str, Any]
    children: dict[str, GraphQLFilterInfo]
    filterset_class: Optional[type[FilterSet]]
